import json
import unittest

import requests_mock
from pandas import DataFrame

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import View, ViewColumn, ColumnType, UserBrief, ViewBrief, QueryDefinition, FilterDefinition, \
    FilterType, Database, Table, Constraints, PrimaryKey, TableBrief, ColumnBrief, Column, ContainerBrief, ImageBrief, \
    Image, Operator
from dbrepo.api.exceptions import ForbiddenError, NotExistsError, MalformedError, AuthenticationError, \
    ResponseCodeError, ExternalSystemError, ServiceError, ServiceConnectionError


class ViewUnitTest(unittest.TestCase):
    database = Database(
        id="6bd39359-b154-456d-b9c2-caa516a45732",
        name='test',
        owner=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='mweise'),
        contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
        exchange_name='dbrepo',
        internal_name='test_abcd',
        is_public=True,
        is_schema_public=True,
        is_dashboard_enabled=True,
        tables=[Table(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                      name="Some Table",
                      description="Test Table",
                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                      internal_name="some_table",
                      owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                      is_versioned=True,
                      queue_name='test',
                      routing_key='dbrepo.test_database_1234.test',
                      is_public=True,
                      is_schema_public=True,
                      constraints=Constraints(uniques=[],
                                              foreign_keys=[],
                                              checks=[],
                                              primary_key=[PrimaryKey(id="1516310f-ecb5-4614-abe2-3b96114e1484",
                                                                      table=TableBrief(
                                                                          id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='Other',
                                                                          internal_name='other',
                                                                          description=None,
                                                                          is_versioned=True,
                                                                          is_public=True,
                                                                          is_schema_public=True,
                                                                          owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                      column=ColumnBrief(
                                                                          id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                                                          table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='id',
                                                                          alias=None,
                                                                          internal_name='id',
                                                                          type=ColumnType.BIGINT))]),
                      columns=[Column(id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                      name="ID",
                                      ord=0,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="id",
                                      type=ColumnType.BIGINT,
                                      is_null_allowed=False),
                               Column(id="85de93a8-834c-4cf4-9d34-f80ebd97e606",
                                      name="Username",
                                      ord=1,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="username",
                                      type=ColumnType.VARCHAR,
                                      is_null_allowed=False)
                               ])],
        container=ContainerBrief(id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                 name='MariaDB Galera 11.1.3',
                                 internal_name='mariadb',
                                 image=ImageBrief(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                  name='mariadb',
                                                  version='11.2.2',
                                                  default=True)))
    image = Image(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                  name='mariadb',
                  version='11.2.2',
                  default=True,
                  data_types=[],
                  operators=[Operator(id="6a96bd99-be3d-4d56-8c38-b14bdfead634",
                                      display_name="IN",
                                      value="IN",
                                      documentation="https://mariadb.com/kb/en/in/")])

    def test_get_views_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/view', json=[])
            # test
            response = RestClient().get_views(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual([], response)

    def test_get_views_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [ViewBrief(id="1b3449d2-780e-4683-9af0-8733e608a4aa",
                             name="Data",
                             internal_name="data",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             initial_view=False,
                             query="SELECT id FROM mytable WHERE deg > 0",
                             query_hash="94c74728b11a690e51d64719868824735f0817b7",
                             owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                             is_public=True,
                             is_schema_public=True)]
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/view', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_views(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_get_views_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/view', status_code=404)
            # test
            try:
                response = RestClient().get_views(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_get_views_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/view', status_code=202)
            # test
            try:
                response = RestClient().get_views(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_get_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = View(id="1b3449d2-780e-4683-9af0-8733e608a4aa",
                       name="Data",
                       internal_name="data",
                       database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                       initial_view=False,
                       query="SELECT id FROM mytable WHERE deg > 0",
                       query_hash="94c74728b11a690e51d64719868824735f0817b7",
                       owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                       is_public=True,
                       is_schema_public=True,
                       columns=[ViewColumn(id="1b3449d2-780e-4683-9af0-8733e608a4aa",
                                           ord=0,
                                           name="id",
                                           internal_name="id",
                                           database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           type=ColumnType.BIGINT,
                                           is_null_allowed=False)],
                       identifiers=[])
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                     json=exp.model_dump())
            # test
            response = RestClient().get_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            self.assertEqual(exp, response)

    def test_get_view_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                     status_code=403)
            # test
            try:
                response = RestClient().get_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                 view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ForbiddenError:
                pass

    def test_get_view_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                     status_code=404)
            # test
            try:
                response = RestClient().get_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                 view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except NotExistsError:
                pass

    def test_get_view_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                     status_code=202)
            # test
            try:
                response = RestClient().get_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                 view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ResponseCodeError:
                pass

    def test_update_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = ViewBrief(id="1b3449d2-780e-4683-9af0-8733e608a4aa",
                            name="Data",
                            internal_name="data",
                            database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                            initial_view=False,
                            query="SELECT id FROM mytable WHERE deg > 0",
                            query_hash="94c74728b11a690e51d64719868824735f0817b7",
                            owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                            is_public=False,
                            is_schema_public=False)
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1', json=exp.model_dump(),
                     status_code=202)
            # test
            response = RestClient(username='foo', password='bar').update_view(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732", view_id=1,
                is_public=False, is_schema_public=False)
            self.assertEqual(exp, response)

    def test_update_view_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1', status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').update_view(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", view_id=1, is_public=False,
                    is_schema_public=False)
            except ForbiddenError:
                pass

    def test_update_view_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1', status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').update_view(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", view_id=1, is_public=False,
                    is_schema_public=False)
            except NotExistsError:
                pass

    def test_update_view_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1', status_code=200)
            # test
            try:
                RestClient(username='foo', password='bar').update_view(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", view_id=1, is_public=False,
                    is_schema_public=False)
            except ResponseCodeError:
                pass

    def test_update_view_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1', status_code=403)
            # test
            try:
                RestClient().update_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", view_id=1, is_public=False,
                                         is_schema_public=False)
            except AuthenticationError:
                pass

    def test_create_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = ViewBrief(id="1b3449d2-780e-4683-9af0-8733e608a4aa",
                            name="Data",
                            internal_name="data",
                            database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                            initial_view=False,
                            query="SELECT id FROM some_table WHERE id IN (1,2)",
                            query_hash="94c74728b11a690e51d64719868824735f0817b7",
                            owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                            is_public=True,
                            is_schema_public=True)
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                          is_public=True, is_schema_public=True,
                                          query=QueryDefinition(datasources=["some_table"],
                                                                columns=["some_table.id"],
                                                                filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                          column="some_table.id",
                                                                                          operator="IN",
                                                                                          value="(1,2)")]))
            self.assertEqual(exp, response)

    def test_create_view_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except MalformedError:
                pass

    def test_create_view_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except ForbiddenError:
                pass

    def test_create_view_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except NotExistsError:
                pass

    def test_create_view_423_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=423)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except ExternalSystemError:
                pass

    def test_create_view_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except ServiceConnectionError:
                pass

    def test_create_view_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except ServiceError:
                pass

    def test_create_view_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                   is_public=True, is_schema_public=True,
                                   query=QueryDefinition(datasources=["some_table"],
                                                         columns=["some_table.id"],
                                                         filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                   column="some_table.id",
                                                                                   operator="IN",
                                                                                   value="(1,2)")]))
            except ResponseCodeError:
                pass

    def test_create_view_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/view', status_code=404)
            # test
            try:
                RestClient().create_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Data",
                                         is_public=True, is_schema_public=True,
                                         query=QueryDefinition(datasources=["some_table"],
                                                               columns=["some_table.id"],
                                                               filters=[
                                                                   FilterDefinition(type=FilterType.WHERE,
                                                                                    column="some_table.id",
                                                                                    operator="IN",
                                                                                    value="(1,2)")]))
            except AuthenticationError:
                pass

    def test_delete_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                               view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")

    def test_delete_view_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except MalformedError:
                pass

    def test_delete_view_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ForbiddenError:
                pass

    def test_delete_view_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except NotExistsError:
                pass

    def test_delete_view_423_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=423)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ExternalSystemError:
                pass

    def test_delete_view_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ServiceConnectionError:
                pass

    def test_delete_view_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ServiceError:
                pass

    def test_delete_view_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                   view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ResponseCodeError:
                pass

    def test_delete_view_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa',
                status_code=403)
            # test
            try:
                RestClient().delete_view(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except AuthenticationError:
                pass

    def test_get_view_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                headers={'X-Headers': 'id,username'}, json=json.dumps(exp))
            # test
            response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                  view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_view_data_dataframe_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(json.dumps(exp))
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                headers={'X-Headers': 'id,username'}, json=json.dumps(exp))
            # test
            response: DataFrame = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            self.assertEqual(df.shape, response.shape)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_view_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=400)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except MalformedError:
                pass

    def test_get_view_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=403)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ForbiddenError:
                pass

    def test_get_view_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=404)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except NotExistsError:
                pass

    def test_get_view_data_409_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=409)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ExternalSystemError:
                pass

    def test_get_view_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=503)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ServiceError:
                pass

    def test_get_view_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=202)
            # test
            try:
                response = RestClient().get_view_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ResponseCodeError:
                pass

    def test_get_view_data_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = 844737
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                headers={'X-Count': str(exp)})
            # test
            response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                        view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            self.assertEqual(exp, response)

    def test_get_view_data_count_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=400)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except MalformedError:
                pass

    def test_get_view_data_count_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=403)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ForbiddenError:
                pass

    def test_get_view_data_count_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=404)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except NotExistsError:
                pass

    def test_get_view_data_count_409_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=409)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ExternalSystemError:
                pass

    def test_get_view_data_count_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=503)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ServiceError:
                pass

    def test_get_view_data_count_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/1b3449d2-780e-4683-9af0-8733e608a4aa/data',
                status_code=202)
            # test
            try:
                response = RestClient().get_view_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                            view_id="1b3449d2-780e-4683-9af0-8733e608a4aa")
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import OntologyBrief


class OntologyUnitTest(unittest.TestCase):

    def test_get_ontologies_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/ontology', json=[])
            # test
            response = RestClient().get_ontologies()
            self.assertEqual([], response)

    def test_get_ontologies_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [OntologyBrief(id="ec6ee082-b6b2-4d65-b931-2f1735f25759",
                                 uri="http://www.ontology-of-units-of-measure.org/resource/om-2/",
                                 prefix="om",
                                 sparql=False,
                                 rdf=True,
                                 uri_pattern="http://www.ontology-of-units-of-measure.org/resource/om-2/.*")]
            # mock
            mock.get('/api/v1/ontology', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_ontologies()
            self.assertEqual(exp, response)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Optional, List

from pydantic import BaseModel


class Datasource(BaseModel):
    uid: str
    type: str
    id: Optional[int] = None
    orgId: Optional[int] = None
    name: Optional[str] = None
    typeLogoUrl: Optional[str] = None
    access: Optional[str] = None
    url: Optional[str] = None
    user: Optional[str] = None
    basicAuth: Optional[bool] = None
    withCredentials: Optional[bool] = None
    isDefault: Optional[bool] = None
    version: Optional[int] = None
    readOnly: Optional[bool] = None
    jsonData: Optional[dict] = None
    secureJsonFields: Optional[dict] = None
    basicAuthUser: Optional[str] = None
    basicAuthPassword: Optional[str] = None
    password: Optional[str] = None


class CreateDatasource(BaseModel):
    name: str
    type: str
    access: str
    url: str
    basicAuth: bool
    version: int
    readOnly: bool
    jsonData: Optional[dict] = None
    secureJsonData: Optional[dict] = None
    basicAuthUser: Optional[str] = None


class Permission(BaseModel):
    permission: str


class CreateDatasourceRequest(BaseModel):
    database_internal_name: str
    readonly: bool
    type: str


class User(BaseModel):
    id: str
    username: str
    roles: List[str]



import os
import unittest

import requests_mock

from dbrepo.RestClient import RestClient


class RestClientUnitTest(unittest.TestCase):

    def test_constructor_succeeds(self):
        # test
        os.environ['REST_API_SECURE'] = 'True'
        response = RestClient()
        self.assertTrue(response.secure)

    def test_constructor_token_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/user', json=[])
            # test
            client = RestClient(password='bar')
            client.get_users()
            self.assertEqual('bar', client.password)

    def test_whoami_anonymous_succeeds(self):
        # test
        response = RestClient().whoami()
        self.assertIsNone(response)

    def test_whoami_succeeds(self):
        # test
        response = RestClient(username="foobar").whoami()
        self.assertEqual("foobar", response)


if __name__ == "__main__":
    unittest.main()



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import ConceptBrief
from dbrepo.api.exceptions import ResponseCodeError


class ContainerUnitTest(unittest.TestCase):

    def test_get_concepts_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [ConceptBrief(id="65586aef-f5b0-446f-b2e1-9dc2a3c0c359",
                                uri="http://dbpedia.org/page/Category:Precipitation",
                                name="Precipitation")]
            # mock
            mock.get('/api/v1/concept', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_concepts()
            self.assertEqual(exp, response)

    def test_get_concepts_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/concept', status_code=202)
            # test
            try:
                RestClient().get_concepts()
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()



"""
The opensearch_client.py is used by the different API endpoints in routes.py to handle requests  to the opensearch db
"""
import logging
import os
from collections.abc import MutableMapping
from json import dumps, load

from opensearchpy import OpenSearch, NotFoundError
from requests import head

from dbrepo.api.dto import Database
from dbrepo.api.exceptions import ForbiddenError, NotExistsError
from dbrepo.core.omlib.constants import OM_IDS
from dbrepo.core.omlib.measure import om
from dbrepo.core.omlib.omconstants import OM
from dbrepo.core.omlib.unit import Unit


class SearchServiceClient:
    """
    The client to communicate with the OpenSearch database.
    """
    instance: OpenSearch = None

    def __init__(self, host: str = None, port: int = 9000, username: str = 'admin', password: str = 'admin'):
        if host is None:
            host = 'search-db'
        self.host = os.getenv('OPENSEARCH_HOST', host)
        self.metadata_endpoint = os.getenv('METADATA_SERVICE_ENDPOINT', 'http://metadata-service:8080')
        self.port = int(os.getenv('OPENSEARCH_PORT', port))
        self.system_username = os.getenv('SYSTEM_USERNAME', username)
        self.system_password = os.getenv('SYSTEM_PASSWORD', password)

    def _instance(self) -> OpenSearch:
        """
        Wrapper method to get the instance singleton.

        @returns: The opensearch instance singleton, if successful.
        """
        if self.instance is None:
            self.instance = OpenSearch(hosts=[{"host": self.host, "port": self.port}],
                                       http_compress=True,
                                       http_auth=(self.system_username, self.system_password))
        return self.instance

    def database_exists(self, database_id: str):
        try:
            SearchServiceClient()._instance().get(index="database", id=database_id)
            return True
        except NotFoundError:
            return False

    def index_update(self, mapping: dict) -> None:
        if SearchServiceClient()._instance().indices.exists(index="database"):
            logging.debug(f"index 'database' exists, removing...")
            SearchServiceClient()._instance().indices.delete(index="database")
        SearchServiceClient()._instance().indices.create(index="database", body=mapping)
        logging.info(f"Created index 'database'")

    def save_database(self, database_id: str, data: Database, fetch: bool = False) -> Database | None:
        """
        Creates the database data with given id. If a document with the id already exists, the document is updated.

        @param database_id: The database id.
        @param data: The database data.
        @param fetch: When enabled, fetches the saved database data. Default: `False`.

        :return The saved database data.

        @returns: The updated database, if successful.
        @throws: opensearchpy.exceptions.NotFoundError If the database was not found in the Search Database. Ignored when `force` is `True`.
        """
        self._instance().update(index="database", id=database_id,
                                body={'doc': data.model_dump(), 'doc_as_upsert': True})
        logging.info(f'Updated database with id: {database_id}')
        if fetch is False:
            return None
        response = self._instance().get(index="database", id=database_id)
        logging.debug(f'fetched database for return value with id: {database_id}')
        return Database.model_validate(response["_source"])

    def delete_database(self, database_id: str) -> None:
        """
        Deletes the database data with given id.

        @param database_id: The database id.
        @throws: opensearchpy.exceptions.NotFoundError If the database was not found in the Search Database.
        """
        self._instance().delete(index="database", id=database_id)
        logging.info(f"Deleted database with id {database_id} in index 'database'")

    def get_fields_for_index(self, field_type: str):
        """
        returns a list of attributes of the data for a specific index.
        :param field_type: The search type
        :return: list of fields
        """
        fields = {
            "database": "*",
            "table": "tables.*",
            "column": "tables.columns.*",
            "concept": "tables.columns.concept.*",
            "unit": "tables.columns.unit.*",
            "identifier": "identifiers.*",
            "view": "views.*",
            "user": "owner.*",
        }
        if field_type not in fields.keys():
            raise NotFoundError(f"Failed to find field type: {field_type}")
        logging.debug(f'requesting field(s) {fields[field_type]} for filter: {field_type}')
        fields = self._instance().indices.get_field_mapping(fields[field_type])
        fields_list = []
        fd = flatten_dict(fields)
        for key in fd.keys():
            if not key.startswith('database'):
                continue
            entry = {}
            if key.split(".")[-1] == "type":
                entry["attr_name"] = key_to_attr_name(key)
                entry["attr_friendly_name"] = attr_name_to_attr_friendly_name(entry["attr_name"])
                entry["type"] = fd[key]
                fields_list.append(entry)
        return fields_list

    def fuzzy_search(self, search_term: str, username: str | None = None, user_token: str | None = None) -> list[
        Database]:
        response = self._instance().search(
            index="database",
            size=1000,
            body={
                "query": {
                    "multi_match": {
                        "query": search_term,
                        "fuzziness": "AUTO",
                        "prefix_length": 2
                    }
                }
            }
        )
        results: list[Database] = []
        if "hits" in response and "hits" in response["hits"]:
            results = [Database.model_validate(hit["_source"]) for hit in response["hits"]["hits"]]
        logging.debug(f'found {len(results)} results')
        return self.filter_results(results, username, user_token)

    def filter_results(self, results: list[Database], username: str | None = None, user_token: str | None = None) -> \
            list[Database]:
        filtered: list[Database] = []
        for database in results:
            if database.is_public or database.is_schema_public:
                logging.debug(f'database with id {database.id} is public or has public schema')
                filtered.append(database)
            elif username is not None and user_token is not None:
                try:
                    url = f'{self.metadata_endpoint}/api/v1/database/{database.id}/access/{username}'
                    logging.debug(f'requesting access from url: {url}')
                    response = head(url=url, auth=(self.system_username, self.system_password))
                    if response.status_code == 200:
                        logging.debug(f'database with id {database.id} is draft and access was found')
                        filtered.append(database)
                    else:
                        logging.warning(
                            f'database with id {database.id} is not accessible: code {response.status_code}')
                except (ForbiddenError, NotExistsError) as e:
                    logging.warning(f'database with id {database.id} is draft but no access was found')
        logging.debug(f'filtered {len(filtered)} results')
        return filtered

    def general_search(self, field_value_pairs: dict = None, username: str | None = None,
                       user_token: str | None = None) -> list[Database]:
        """
        Main method for searching stuff in the opensearch db

        all parameters are optional

        :param field_type: The index to be searched. Optional.
        :param field_value_pairs: The key-value pair of properties that need to match. Optional.
        :return: The object of results and HTTP status code. e.g. { "hits": { "hits": [] } }, 200
        """
        musts = []
        if field_value_pairs is not None and len(field_value_pairs) > 0:
            logging.debug(f'field_value_pairs present: {field_value_pairs}')
            for key, value in field_value_pairs.items():
                if field_value_pairs[key] == None:
                    logging.debug(f"skip empty key: {key}")
                    continue
                logging.debug(f"processing key: {key}")
                if '.' in key:
                    logging.debug(f'key {key} is nested: use nested query')
                    musts.append({
                        "match": {
                            key: value
                        }
                    })
                else:
                    logging.debug(f'key {key} is flat: use bool query')
                    musts.append({
                        "match": {
                            key: {"query": value, "minimum_should_match": "90%"}
                        }
                    })
        body = {
            "query": {"bool": {"must": musts}}
        }
        logging.debug(f'search: {body}')
        response = self._instance().search(
            index="database",
            size=1000,
            body=dumps(body)
        )
        results: list[Database] = []
        if "hits" in response and "hits" in response["hits"]:
            results = [Database.model_validate(hit["_source"]) for hit in response["hits"]["hits"]]
        logging.debug(f'found {len(results)} results')
        return self.filter_results(results, username, user_token)

    def unit_independent_search(self, t1: float, t2: float, field_value_pairs: dict, username: str | None = None) -> \
            list[Database]:
        """
        Main method for searching stuff in the opensearch db

        :param t1: start value
        :param t2: end value
        :param field_value_pairs: the key-value pairs, optional.
        :param username: the username, optional.
        :return:
        """
        logging.info(f"Performing unit-independent search")
        searches = []
        body = {
            "size": 0,
            "aggs": {
                "units": {
                    "terms": {"field": "unit.uri", "size": 500}
                }
            }
        }
        response = self._instance().search(
            index="database",
            size=1000,
            body=dumps(body)
        )
        unit_uris = [hit["key"] for hit in response["aggregations"]["units"]["buckets"]]
        logging.debug(f"found {len(unit_uris)} unit(s) in column index")
        if len(unit_uris) == 0:
            raise NotFoundError("Failed to search: no unit assigned")
        base_unit = unit_uri_to_unit(field_value_pairs["unit.uri"])
        for unit_uri in unit_uris:
            gte = t1
            lte = t2
            if unit_uri != field_value_pairs["unit.uri"]:
                target_unit = unit_uri_to_unit(unit_uri)
                if not Unit.can_convert(base_unit, target_unit):
                    logging.error(f"Cannot convert unit {field_value_pairs['unit.uri']} to target unit {unit_uri}")
                    continue
                gte = om(t1, base_unit).convert(target_unit)
                lte = om(t2, base_unit).convert(target_unit)
                logging.debug(
                    f"converted original range [{t1},{t2}] for base unit {base_unit} to mapped range [{gte},{lte}] for target unit={target_unit}")
            searches.append({'index': 'column'})
            searches.append({
                "query": {
                    "bool": {
                        "must": [
                            {
                                "match": {
                                    "concept.uri": {
                                        "query": field_value_pairs["concept.uri"]
                                    }
                                }
                            },
                            {
                                "range": {
                                    "val_min": {
                                        "gte": gte
                                    }
                                }
                            },
                            {
                                "range": {
                                    "val_max": {
                                        "lte": lte
                                    }
                                }
                            },
                            {
                                "match": {
                                    "unit.uri": {
                                        "query": unit_uri
                                    }
                                }
                            }
                        ]
                    }
                }
            })
        logging.debug('searches: %s', searches)
        body = ''
        for search in searches:
            body += '%s \n' % dumps(search)
        response = self._instance().msearch(
            body=dumps(body)
        )
        results = flatten([hits["hits"]["hits"] for hits in response["responses"]])
        return [database for database in results if
                database.is_public or database.is_schema_public or (
                        username is not None and database.owner.username == username)]


def key_to_attr_name(key: str) -> str:
    """
    Maps an attribute key to a machine-readable representation
    :param key: The attribute key
    :return: The machine-readable representation of the attribute key
    """
    parts = []
    previous = None
    for part in key.split(".")[1:-1]:  # remove the first and last sub-item database.xxx.yyy.zzz.type -> xxx.yyy.zzz
        if part == "mappings" or part == "mapping":  # remove the mapping sub-item(s)
            continue
        if part == previous:  # remove redundant sub-item(s)
            continue
        previous = part
        parts.append(part)
    return ".".join(parts)


def attr_name_to_attr_friendly_name(key: str) -> str:
    """
    Maps an attribute key to a human-readable representation
    :param key: The attribute key
    :return: The human-readable representation of the attribute key
    """
    with open('./friendly_names_overrides.json') as json_data:
        d = load(json_data)
        for json_key in d.keys():
            if json_key == key:
                logging.debug(f"friendly name exists for key {json_key}")
                return d[json_key]
    return ''.join(key.replace('_', ' ').title().split('.')[-1:])


def flatten_dict(
        d: MutableMapping, parent_key: str = "", sep: str = "."
) -> MutableMapping:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def flatten(mylist):
    return [item for sublist in mylist for item in sublist]


def unit_uri_to_unit(uri):
    base_identifier = uri[len(OM_IDS.NAMESPACE):].replace("-", "")
    return getattr(OM, base_identifier)



from rdflib import URIRef

from dbrepo.core.omlib.dimension import Dimension
from dbrepo.core.omlib.exceptions.dimensionexception import DimensionalException
from dbrepo.core.omlib.exceptions.unitconversionexception import ScaleConversionException
from dbrepo.core.omlib.exceptions.unitidentityexception import ScaleIdentityException
from dbrepo.core.omlib.thing import Thing
from dbrepo.core.omlib.unit import Unit


class Scale(Thing):
    _scales = []

    @staticmethod
    def clear_cache():
        Scale._scales.clear()

    @staticmethod
    def _add_when_not_in_cache(scale):
        for test_scale in Scale._scales:
            if test_scale == scale:
                if test_scale.identifier == scale.identifier or not isinstance(scale.identifier, URIRef) \
                        or not isinstance(test_scale.identifier, URIRef):
                    if isinstance(test_scale, RatioScale) and isinstance(scale, RatioScale):
                        if test_scale.unit == scale.unit:
                            if isinstance(scale.identifier, URIRef) and not isinstance(test_scale.identifier, URIRef):
                                test_scale.identifier = scale.identifier
                            return test_scale
                    if isinstance(test_scale, IntervalScale) and isinstance(scale, IntervalScale):
                        if test_scale.unit == scale.unit and test_scale.baseScale == scale.baseScale \
                                and test_scale.offSet == scale.offSet:
                            if isinstance(scale.identifier, URIRef) and not isinstance(test_scale.identifier, URIRef):
                                test_scale.identifier = scale.identifier
                            return test_scale
                    if test_scale.identifier == scale.identifier:
                        raise ScaleIdentityException("The scale with identifier {} already exists but with different"
                                                     "properties. \n{} \nis not the same as \n{}".format(scale.identifier,
                                                     scale, test_scale))
        Scale._scales.append(scale)
        return scale

    @staticmethod
    def with_label(label):
        return_scales = []
        for test_scale in Scale._scales:
            for test_label in test_scale.allLabels:
                if test_label == label:
                    return_scales.append(test_label)
        return return_scales

    @staticmethod
    def with_identifier(identifier):
        for test_scale in Scale._scales:
            if test_scale.identifier == identifier:
                return test_scale
        return None

    @staticmethod
    def with_dimensions(dimensions, in_system_of_units=None):
        return_scales = []
        for test_scale in Scale._scales:
            if test_scale.dimensions == dimensions and test_scale.systemOfUnits == in_system_of_units:
                return_scales.append(test_scale)
        return return_scales

    @staticmethod
    def get_ratio_scale(unit, label=None, identifier=None):
        scale = RatioScale(unit, label, identifier)
        scale = Scale._add_when_not_in_cache(scale)
        return scale

    @staticmethod
    def get_interval_scale(base_scale, unit, off_set, label=None, identifier=None):
        scale = IntervalScale(base_scale, unit, off_set, label, identifier)
        scale = Scale._add_when_not_in_cache(scale)
        return scale

    @staticmethod
    def conversion_factor(from_scale, to_scale):
        if (isinstance(from_scale, RatioScale) or isinstance(from_scale, IntervalScale)) and \
                (isinstance(to_scale, RatioScale) or isinstance(to_scale, IntervalScale)):
            can_convert = Unit.can_convert(from_scale.unit, to_scale.unit)
            if not can_convert:
                raise DimensionalException("A scale with dimensions {} cannot be converted to a scale with "
                                           "dimensions {}."
                                           .format(from_scale.dimensions, to_scale.dimensions))
            factor = Unit.conversion_factor(from_scale.unit, to_scale.unit)
            return factor
        else:
            raise ScaleConversionException("Cannot convert from {} to {} as they are not both cardinal scales."
                                            .format(from_scale, to_scale))

    @staticmethod
    def conversion_off_set(from_scale, to_scale):
        if (isinstance(from_scale, RatioScale) or isinstance(from_scale, IntervalScale)) and \
                (isinstance(to_scale, RatioScale) or isinstance(to_scale, IntervalScale)):
            from_ratio_scale = from_scale.base_ratio_scale()
            to_ratio_scale = to_scale.base_ratio_scale()
            from_ratio_factor = Unit.conversion_factor(from_scale.unit,from_ratio_scale[0].unit)
            to_ratio_factor = Unit.conversion_factor(to_scale.unit,to_ratio_scale[0].unit)
            if from_ratio_scale[0] == to_ratio_scale[0]:
                off_set = (to_ratio_scale[1] * to_ratio_factor - from_ratio_scale[1]*from_ratio_factor) \
                          / to_ratio_factor
                return off_set
            else:
                raise ScaleConversionException("Cannot convert from {} to {} as they do not use the same base ratio "
                                               "scale, i.e. they do not have the same known zero point."
                                               .format(from_scale, to_scale))
        else:
            raise ScaleConversionException("Cannot convert from {} to {} as they are not both cardinal scales."
                                            .format(from_scale, to_scale))

    def __init__(self, label=None, identifier=None, dimensions=Dimension(), system_of_units=None):
        super().__init__(label, identifier)
        self.fixedPoints = []
        self.dimensions = dimensions
        self.systemOfUnits = system_of_units

    def add_fixed_point(self, fixed_point):
        self.fixedPoints.append(fixed_point)

    def __str__(self):
        return f'{self.label()}\t<{self.identifier}>  dim: {self.dimensions}'

    def __eq__(self, other):
        if isinstance(other, Scale):
            return str(self.identifier) == str(other.identifier)
        return False

    def __ne__(self, other):
        return not (self == other)


class RatioScale(Scale):

    def __init__(self, unit, label=None, identifier=None, system_of_units=None):
        if not isinstance(unit, Unit):
            raise ValueError("The unit parameter in RatioScale is required to be of type Unit.")
        if system_of_units is None:
            system_of_units = unit.systemOfUnits
        super().__init__(label, identifier, dimensions=unit.dimensions, system_of_units=system_of_units)
        self.unit = unit

    def base_ratio_scale(self):
        return self, 0.0

    def __eq__(self, other):
        if isinstance(other, RatioScale):
            if str(self.identifier) == str(other.identifier):
                return True
            if isinstance(self.identifier, URIRef) and isinstance(other.identifier, URIRef):
                return False
            if self.unit == other.unit:
                return True
        return False

    def __str__(self):
        return f'{self.label()}\t<{self.identifier}> unit: {self.unit} dim: {self.dimensions}'


class IntervalScale(Scale):

    def __init__(self, base_scale, unit, off_set, label=None, identifier=None, system_of_units=None):
        if not isinstance(unit, Unit):
            raise ValueError("The unit parameter in IntervalScale is required to be of type Unit.")
        if not isinstance(base_scale, Scale):
            raise ValueError("The base_scale parameter in IntervalScale is required to be of type Scale.")
        if unit.dimensions != base_scale.dimensions:
            raise DimensionalException("The dimensions of the base scale are not the same as the dimensions of the"
                                       " unit used. {} is not the same as {}"
                                       .format(base_scale.dimensions, unit.dimensions))
        if system_of_units is None:
            system_of_units = base_scale.systemOfUnits
        if system_of_units is None:
            system_of_units = unit.systemOfUnits
        super().__init__(label, identifier, dimensions=unit.dimensions, system_of_units=system_of_units)
        self.unit = unit
        self.baseScale = base_scale
        self.offSet = off_set

    def base_ratio_scale(self):
        base_base = self.baseScale.base_ratio_scale()
        conversion_fac = Unit.conversion_factor(base_base[0].unit, self.unit)
        return base_base[0], base_base[1] * conversion_fac + self.offSet

    def __eq__(self, other):
        if isinstance(other, IntervalScale):
            if str(self.identifier) == str(other.identifier):
                return True
            if self.unit == other.unit and self.baseScale == other.baseScale and self.offSet == other.offSet:
                return True
        return False

    def __str__(self):
        return f'{self.label()}\t<{self.identifier}> base: {self.baseScale} unit: {self.unit} dim: {self.dimensions}'



#!/usr/bin/env python3
import os
from distutils.core import setup

app_version = os.environ.get("APP_VERSION", "1.13.1")
doc_version = os.environ.get("DOC_VERSION", "1.13")

setup(name="dbrepo",
      version=str(app_version),
      description="A library for communicating with DBRepo",
      url=f"https://www.ifs.tuwien.ac.at/infrastructures/dbrepo/{doc_version}/",
      author="Martin Weise",
      license="Apache-2.0",
      author_email="martin.weise@tuwien.ac.at",
      packages=[
          "dbrepo",
          "dbrepo.api",
          "dbrepo.core",
          "dbrepo.core.api",
          "dbrepo.core.client",
          "dbrepo.core.omlib",
          "dbrepo.core.omlib.exceptions",
          "dbrepo.core.omlib.rdf",
      ])



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import ImageBrief


class ImageUnitTest(unittest.TestCase):

    def test_get_images_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/image', json=[])
            # test
            response = RestClient().get_images()
            self.assertEqual([], response)

    def test_get_images_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [ImageBrief(id="96c1876a-7473-44fd-8115-19ca6fde32d4",
                              name="mariadb",
                              version="11.1.3",
                              default=False)]
            # mock
            mock.get('/api/v1/image', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_images()
            self.assertEqual(exp, response)


if __name__ == "__main__":
    unittest.main()



import datetime
import json
import unittest

import requests_mock
from pandas import DataFrame

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Table, Column, Constraints, ColumnType, ConceptBrief, UnitBrief, \
    TableStatistics, ColumnStatistic, PrimaryKey, ColumnBrief, TableBrief, UserBrief, History, HistoryEventType
from dbrepo.api.exceptions import MalformedError, ForbiddenError, NotExistsError, NameExistsError, \
    AuthenticationError, ExternalSystemError, ServiceError, ServiceConnectionError, ResponseCodeError


class TableUnitTest(unittest.TestCase):

    def test_create_table_succeeds(self):
        exp = TableBrief(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                         database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                         name="Test",
                         description="Test Table",
                         internal_name="test",
                         owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                         is_versioned=True,
                         is_public=True,
                         is_schema_public=True)
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', json=exp.model_dump(),
                      headers={'X-Headers': 'id,name'}, status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                           description="Test Table", dataframe=dataframe, with_data=False,
                                           is_public=True, is_schema_public=True)
            self.assertEqual(exp, response)

    def test_create_table_index_missing_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}])
            # test
            client = RestClient(username="a", password="b")
            try:
                client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                    description="Test Table", dataframe=dataframe, with_data=False,
                                    is_public=True, is_schema_public=True)
            except MalformedError:
                pass

    def test_create_table_400_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except MalformedError:
                pass

    def test_create_table_403_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=403)
            # test
            try:
                RestClient(username="a", password="b").create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                    name="Test",
                                                                    description="Test Table", dataframe=dataframe,
                                                                    is_public=True, is_schema_public=True)
            except ForbiddenError:
                pass

    def test_create_table_404_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except NotExistsError:
                pass

    def test_create_table_409_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=409)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except NameExistsError:
                pass

    def test_create_table_502_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except ServiceConnectionError:
                pass

    def test_create_table_503_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except ServiceError:
                pass

    def test_create_table_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                               description="Test Table", dataframe=dataframe,
                                               is_public=True, is_schema_public=True)
            except ResponseCodeError:
                pass

    def test_create_table_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            dataframe = DataFrame.from_records([{'id': 1, 'name': 'foobar'}], index=['id'])
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=409)
            # test
            try:
                response = RestClient().create_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732", name="Test",
                                                     description="Test Table", dataframe=dataframe,
                                                     is_public=True, is_schema_public=True)
            except AuthenticationError:
                pass

    def test_get_tables_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', json=[])
            # test
            response = RestClient().get_tables(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual([], response)

    def test_get_tables_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [TableBrief(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                              name="Test",
                              description="Test Table",
                              database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                              internal_name="test",
                              is_public=True,
                              is_schema_public=True,
                              owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                              is_versioned=True)]
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_tables(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_get_tables_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=403)
            # test
            try:
                RestClient().get_tables(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError:
                pass

    def test_get_tables_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=404)
            # test
            try:
                RestClient().get_tables(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_get_tables_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table', status_code=202)
            # test
            try:
                RestClient().get_tables(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_get_table_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Table(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                        name="Test",
                        description="Test Table",
                        database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                        internal_name="test",
                        owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                        is_versioned=True,
                        queue_name='test',
                        routing_key='dbrepo.test_database_1234.test',
                        is_public=True,
                        is_schema_public=True,
                        constraints=Constraints(uniques=[],
                                                foreign_keys=[],
                                                checks=[],
                                                primary_key=[PrimaryKey(id="1516310f-ecb5-4614-abe2-3b96114e1484",
                                                                        table=TableBrief(
                                                                            id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                            database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                            name='Other',
                                                                            internal_name='other',
                                                                            description=None,
                                                                            is_versioned=True,
                                                                            is_public=True,
                                                                            is_schema_public=True,
                                                                            owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                        column=ColumnBrief(
                                                                            id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                                                            table_id="",
                                                                            database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                            name='id',
                                                                            alias=None,
                                                                            internal_name='id',
                                                                            type=ColumnType.BIGINT))]),
                        columns=[Column(id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                        name="ID",
                                        ord=0,
                                        database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                        internal_name="id",
                                        type=ColumnType.BIGINT,
                                        is_null_allowed=False)])
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                     json=exp.model_dump())
            # test
            response = RestClient().get_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                              table_id="b3230b86-4743-498d-9015-3fad58049692")
            self.assertEqual(exp, response)

    def test_get_table_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                     status_code=403)
            # test
            try:
                response = RestClient().get_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                  table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ForbiddenError:
                pass

    def test_get_table_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                     status_code=404)
            # test
            try:
                response = RestClient().get_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                  table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_get_table_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                     status_code=202)
            # test
            try:
                response = RestClient().get_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                  table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass

    def test_delete_table_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                table_id="b3230b86-4743-498d-9015-3fad58049692")

    def test_delete_table_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except MalformedError:
                pass

    def test_delete_table_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ForbiddenError:
                pass

    def test_delete_table_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_delete_table_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceConnectionError:
                pass

    def test_delete_table_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceError:
                pass

    def test_delete_table_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass

    def test_delete_table_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692',
                        status_code=404)
            # test
            try:
                RestClient().delete_table(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                          table_id="b3230b86-4743-498d-9015-3fad58049692")
            except AuthenticationError:
                pass

    def test_get_table_history_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [History(event=HistoryEventType.INSERT,
                           total=2,
                           timestamp=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc))]
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                json=[exp[0].model_dump()])
            # test
            response = RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            self.assertEqual(exp, response)

    def test_get_table_history_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                status_code=400)
            # test
            try:
                RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692")
            except MalformedError:
                pass

    def test_get_table_history_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                status_code=403)
            # test
            try:
                RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ForbiddenError:
                pass

    def test_get_table_history_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                status_code=404)
            # test
            try:
                RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_get_table_history_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                status_code=503)
            # test
            try:
                RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceError:
                pass

    def test_get_table_history_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/history?size=100',
                status_code=202)
            # test
            try:
                RestClient().get_table_history(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass

    def test_get_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                headers={'X-Headers': 'id,username'}, json=exp)
            # test
            response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   table_id="b3230b86-4743-498d-9015-3fad58049692",
                                                   timestamp=datetime.datetime(2024, 1, 1, 0, 0, 0, 0,
                                                                               datetime.timezone.utc))
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_table_data_dataframe_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                headers={'X-Headers': 'id,username'}, json=exp)
            # test
            response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   table_id="b3230b86-4743-498d-9015-3fad58049692")
            self.assertEqual(df.shape, response.shape)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_table_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=400)
            # test
            try:
                response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                       table_id="b3230b86-4743-498d-9015-3fad58049692")
            except MalformedError:
                pass

    def test_get_table_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=403)
            # test
            try:
                response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                       table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ForbiddenError:
                pass

    def test_get_table_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                       table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_get_table_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                       table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceError:
                pass

    def test_get_table_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=202)
            # test
            try:
                response = RestClient().get_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                       table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass

    def test_get_table_data_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = 2
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                headers={'X-Count': str(exp)})
            # test
            response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                                         timestamp=datetime.datetime(2024, 1, 1, 0, 0, 0, 0))
            self.assertEqual(exp, response)

    def test_get_table_data_count_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=400)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except MalformedError:
                pass

    def test_get_table_data_count_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=403)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ForbiddenError:
                pass

    def test_get_table_data_count_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_get_table_data_count_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceError:
                pass

    def test_get_table_data_count_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=202)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass

    def test_get_table_data_count_not_countable_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=409)
            # test
            try:
                response = RestClient().get_table_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ExternalSystemError:
                pass

    def test_create_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=201)
            # test
            client = RestClient(username="a", password="b")
            client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     table_id="b3230b86-4743-498d-9015-3fad58049692",
                                     data={'name': 'Josiah', 'age': 45, 'gender': 'male'})

    def test_create_table_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except MalformedError:
                pass

    def test_create_table_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except ForbiddenError:
                pass

    def test_create_table_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except NotExistsError:
                pass

    def test_create_table_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except ServiceError:
                pass

    def test_create_table_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except ResponseCodeError:
                pass

    def test_create_table_data_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                RestClient().create_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692",
                                               data={'name': 'Josiah', 'age': 45, 'gender': 'male'})
            except AuthenticationError:
                pass

    def test_update_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     table_id="b3230b86-4743-498d-9015-3fad58049692",
                                     data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                     keys={'id': 1})

    def test_update_table_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except MalformedError:
                pass

    def test_update_table_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except ForbiddenError:
                pass

    def test_update_table_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except NotExistsError:
                pass

    def test_update_table_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except ServiceError:
                pass

    def test_update_table_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                                         data={'name': 'Josiah', 'age': 45, 'gender': 'male'},
                                         keys={'id': 1})
            except ResponseCodeError:
                pass

    def test_delete_table_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})

    def test_delete_table_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})
            except MalformedError:
                pass

    def test_delete_table_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})
            except ForbiddenError:
                pass

    def test_delete_table_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})
            except NotExistsError:
                pass

    def test_delete_table_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})
            except ServiceError:
                pass

    def test_delete_table_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                         table_id="b3230b86-4743-498d-9015-3fad58049692", keys={'id': 1})
            except ResponseCodeError:
                pass

    def test_delete_table_data_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/data',
                status_code=404)
            # test
            try:
                RestClient().delete_table_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               table_id="b3230b86-4743-498d-9015-3fad58049692",
                                               keys={'id': 1})
            except AuthenticationError:
                pass

    def test_update_table_column_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Column(id="2f30858d-b47b-4068-9028-fb0524ddf6cb",
                         ord=0,
                         name="ID",
                         database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                         table_id="b3230b86-4743-498d-9015-3fad58049692",
                         internal_name="id",
                         type=ColumnType.BIGINT,
                         concept=ConceptBrief(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                              uri="http://dbpedia.org/page/Category:Precipitation",
                                              name="Precipitation"),
                         unit=UnitBrief(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                        uri="http://www.wikidata.org/entity/Q119856947",
                                        name="liters per square meter"),
                         is_null_allowed=False)
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                json=exp.model_dump(),
                status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                  table_id="b3230b86-4743-498d-9015-3fad58049692",
                                                  column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                                  unit_uri="http://www.wikidata.org/entity/Q119856947",
                                                  concept_uri="http://dbpedia.org/page/Category:Precipitation")
            self.assertEqual(exp, response)

    def test_update_table_column_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except MalformedError:
                pass

    def test_update_table_column_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except ForbiddenError:
                pass

    def test_update_table_column_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except NotExistsError:
                pass

    def test_update_table_column_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except ServiceConnectionError:
                pass

    def test_update_table_column_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except ServiceError:
                pass

    def test_update_table_column_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                           table_id="b3230b86-4743-498d-9015-3fad58049692",
                                           column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                           unit_uri="http://www.wikidata.org/entity/Q119856947",
                                           concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except ResponseCodeError:
                pass

    def test_update_table_column_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/column/758da021-f809-4d87-a866-9d4664a039dc',
                status_code=404)
            # test
            try:
                RestClient().update_table_column(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                 table_id="b3230b86-4743-498d-9015-3fad58049692",
                                                 column_id="758da021-f809-4d87-a866-9d4664a039dc",
                                                 unit_uri="http://www.wikidata.org/entity/Q119856947",
                                                 concept_uri="http://dbpedia.org/page/Category:Precipitation")
            except AuthenticationError:
                pass

    def test_analyse_table_statistics_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = TableStatistics(
                total_columns=1,
                total_rows=1000,
                columns={
                    "id": ColumnStatistic(name="id", val_min=1.0, val_max=9.0, mean=5.0, median=5.0, std_dev=2.73)})
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                json=exp.model_dump(), status_code=202)
            # test
            response = RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                             table_id="b3230b86-4743-498d-9015-3fad58049692")
            self.assertEqual(exp, response)

    def test_analyse_table_statistics_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                status_code=400)
            # test
            try:
                RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            except MalformedError:
                pass

    def test_analyse_table_statistics_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                status_code=404)
            # test
            try:
                RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            except NotExistsError:
                pass

    def test_analyse_table_statistics_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                status_code=502)
            # test
            try:
                RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceConnectionError:
                pass

    def test_analyse_table_statistics_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                status_code=503)
            # test
            try:
                RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ServiceError:
                pass

    def test_analyse_table_statistics_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/analyse/database/6bd39359-b154-456d-b9c2-caa516a45732/table/b3230b86-4743-498d-9015-3fad58049692/statistics',
                status_code=200)
            # test
            try:
                RestClient().analyse_table_statistics(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                      table_id="b3230b86-4743-498d-9015-3fad58049692")
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Container, Image, ContainerBrief, ImageBrief, DataType, Operator
from dbrepo.api.exceptions import ResponseCodeError, NotExistsError, AuthenticationError, MalformedError, \
    ForbiddenError, NameExistsError


class ContainerUnitTest(unittest.TestCase):

    def test_create_container_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Container(id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                            name="MariaDB 10.11.3",
                            internal_name="mariadb_10_11_3",
                            image=Image(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                        name="mariadb",
                                        version="10.11.3",
                                        default=True,
                                        operators=[
                                            Operator(id="0917b17e-5d68-4ddf-94f6-f178f74a0dae",
                                                     display_name="XOR",
                                                     value="XOR",
                                                     documentation="https://mariadb.com/kb/en/xor/")],
                                        data_types=[
                                            DataType(id="22975809-5496-4d67-9fd4-6689f0030f82",
                                                     display_name="SERIAL",
                                                     value="serial",
                                                     documentation="https://mariadb.com/kb/en/bigint/",
                                                     is_quoted=False,
                                                     is_buildable=True)]))
            # mock
            mock.post('/api/v1/container', json=exp.model_dump(), status_code=201)
            # test
            response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                   host='data-db2',
                                                                                   image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                   privileged_username='root',
                                                                                   privileged_password='dbrepo',
                                                                                   port=3306)
            self.assertEqual(exp, response)

    def test_create_container_anonymous_fails(self):
        # test
        try:
            response = RestClient().create_container(name='MariaDB 10.11.3', host='data-db2',
                                                     image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                     privileged_username='root', privileged_password='dbrepo',
                                                     port=3306)
        except AuthenticationError:
            pass

    def test_create_container_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/container', status_code=400)
            # test
            try:
                response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                       host='data-db2',
                                                                                       image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                       privileged_username='root',
                                                                                       privileged_password='dbrepo',
                                                                                       port=3306)
            except MalformedError:
                pass

    def test_create_container_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/container', status_code=403)
            # test
            try:
                response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                       host='data-db2',
                                                                                       image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                       privileged_username='root',
                                                                                       privileged_password='dbrepo',
                                                                                       port=3306)
            except ForbiddenError:
                pass

    def test_create_container_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/container', status_code=404)
            # test
            try:
                response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                       host='data-db2',
                                                                                       image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                       privileged_username='root',
                                                                                       privileged_password='dbrepo',
                                                                                       port=3306)
            except NotExistsError:
                pass

    def test_create_container_409_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/container', status_code=409)
            # test
            try:
                response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                       host='data-db2',
                                                                                       image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                       privileged_username='root',
                                                                                       privileged_password='dbrepo',
                                                                                       port=3306)
            except NameExistsError:
                pass

    def test_create_container_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/container', status_code=200)
            # test
            try:
                response = RestClient(username="foo", password="bar").create_container(name='MariaDB 10.11.3',
                                                                                       host='data-db2',
                                                                                       image_id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                                                       privileged_username='root',
                                                                                       privileged_password='dbrepo',
                                                                                       port=3306)
            except ResponseCodeError:
                pass

    def test_get_containers_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/container', json=[])
            # test
            response = RestClient().get_containers()
            self.assertEqual([], response)

    def test_get_containers_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [
                ContainerBrief(id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                               name="MariaDB 10.11.3",
                               internal_name="mariadb_10_11_3",
                               running=True,
                               image=ImageBrief(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                name="mariadb",
                                                version="10.11.3",
                                                default=True),
                               hash="f829dd8a884182d0da846f365dee1221fd16610a14c81b8f9f295ff162749e50")
            ]
            # mock
            mock.get('/api/v1/container', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_containers()
            self.assertEqual(exp, response)

    def test_get_containers_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/container', status_code=204)
            # test
            try:
                response = RestClient().get_containers()
            except ResponseCodeError:
                pass

    def test_get_container_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Container(id="",
                            name="MariaDB 10.11.3",
                            internal_name="mariadb_10_11_3",
                            image=Image(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                        name="mariadb",
                                        version="10.11.3",
                                        default=True,
                                        operators=[
                                            Operator(id="0917b17e-5d68-4ddf-94f6-f178f74a0dae",
                                                     display_name="XOR",
                                                     value="XOR",
                                                     documentation="https://mariadb.com/kb/en/xor/")],
                                        data_types=[
                                            DataType(id="22975809-5496-4d67-9fd4-6689f0030f82",
                                                     display_name="SERIAL",
                                                     value="serial",
                                                     documentation="https://mariadb.com/kb/en/bigint/",
                                                     is_quoted=False,
                                                     is_buildable=True)]))
            # mock
            mock.get('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', json=exp.model_dump())
            # test
            response = RestClient().get_container(container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            self.assertEqual(exp, response)

    def test_get_container_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=404)
            # test
            try:
                response = RestClient().get_container(container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except NotExistsError:
                pass

    def test_get_container_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=401)
            # test
            try:
                response = RestClient().get_container(container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except ResponseCodeError:
                pass

    def test_delete_container_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=202)
            # test
            RestClient(username='foo', password='bar').delete_container(
                container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")

    def test_delete_container_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=400)
            # test
            try:
                RestClient(username='foo', password='bar').delete_container(
                    container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except MalformedError:
                pass

    def test_delete_container_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').delete_container(
                    container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except ForbiddenError:
                pass

    def test_delete_container_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').delete_container(
                    container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except NotExistsError:
                pass

    def test_delete_container_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/v1/container/44d811a8-4019-46ba-bd57-ea10a2eb0c74', status_code=200)
            # test
            try:
                RestClient(username='foo', password='bar').delete_container(
                    container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
            except ResponseCodeError:
                pass

    def test_delete_container_anonymous_fails(self):
        # test
        try:
            RestClient().delete_container(container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74")
        except AuthenticationError:
            pass


if __name__ == "__main__":
    unittest.main()
